# edl : common library for the energy-dashboard tool-chain
# Copyright (C) 2019  Todd Greenwood-Geer (Enviro Software Solutions, LLC)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
filesystem.py : compute things like filenames
"""

import os
import pdb

BAD_S3_CHARS = ['&', '@', ':', ',', '$', '=', '+', '?', ';', ' ', '\\', '^', '>', '<', '{', '}', '[', ']', '%', '~', '|']
#PROTOCOLS = ["http://", "https://", "ftp://"]
# TODO: yes, there are replacement chars that are removed in the linter
# TODO: need to sort this out after the migration
COMMON_REPLACEMENTS = [
    ("http://oasis.caiso.com/oasisapi/", "oasis_"),
    ("http://zwrob.com/", ""),
    ("/", "_"),
    ("&", ","),
    ("SingleZip", "SZ"),
    ("queryname", "q"),
    ("startdatetime", "sdt"),
    ("enddatetime", "edt"),
    ("market_run_id", "mri"),
    ("version", "v")
]

def glob_dir(path, ending):
    """Returns files matching ending (simple glob), does NOT DO REGEX"""
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file() and entry.name.lower().endswith(ending.lower()):
                yield(entry.name)

def shrink_file_name(name, rtuples):
    """Shrink file name using provided replacements in rtuples"""
    for (a,b) in rtuples:
        name = name.replace(a,b)
    return name

def s3lint_file_name(name):
    """Replace characters known to break S3"""
    for c in BAD_S3_CHARS:
        name = name.replace(c, "_")
    return name

def url2filename(url, rtuples = COMMON_REPLACEMENTS, ending = ".zip"):
    """
    Download URLs may contain characters that are not OK on either
    the local filesystem or the remote store (S3). `url2filename`
    returns a filename that is safe for either.

    url     : url for the downloaded resource
    ending  : file name ending, like '.zip' for the created local file
    rtuples : list of (string, string) tuple replacements 
    """
    name = s3lint_file_name(shrink_file_name(url, rtuples))
    return "%s%s" % (name, ending)

def clean_legacy_filename(name, ending=".zip"):
    """
    Convert legacy file name to a properly linted name. 
    Note: name should end with 'ending' 
    """
    if name.endswith(ending):
        return "%s%s" % (s3lint_file_name(name[:-len(ending)]), ending)
    else:
        raise Exception("bad file name: expected name to end with '%s'; name: %s" % (ending, name))



# edl : common library for the energy-dashboard tool-chain
# Copyright (C) 2019  Todd Greenwood-Geer (Enviro Software Solutions, LLC)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import logging
import sqlite3
from edl.resources import log
from edl.resources import filesystem

class MemDb():
    def __init__(self, db_path):
        self.db_path = db_path
        self.filedb = None
        self.memdb = None
    def open(self):
        # source
        source = self.filedb = sqlite3.connect(self.db_path)
        # dest
        dest = self.memdb = sqlite3.connect(':memory:')
        # backup
        source.backup(dest)
        return dest
    def close(self):
        # source
        source = self.memdb
        # dest
        dest = self.filedb
        # backup
        source.backup(dest)
        return dest
    def db(self):
        return self.memdb
    def __repr__(self):
        return str(self.db_path)

class DbMgr():
    def __init__(self, logger, resource_name):
        self.dbs = {}
        self.resource_name = resource_name
        self.logger = logger
    def get(self, db_path):
        log.debug(self.logger, {
            "name"      : __name__,
            "src"       : self.resource_name,
            "method"    : "DbMgr.get",
            "db_path"   : db_path,
            })
        if db_path not in self.dbs:
            db = MemDb(db_path)
            cnx = db.open()
            self.dbs[db_path] = db
            log.debug(self.logger, {
                "name"      : __name__,
                "src"       : self.resource_name,
                "method"    : "DbMgr.get",
                "db_path"   : db_path,
                "message"   : "Created in memory db",
                })
        return self.dbs[db_path].db()
    def __enter__(self):
        return self
    def __exit__(self, type, value, traceback):
        for k,v in self.dbs.items():
            v.close()
            log.debug(self.logger, {
                "name"      : __name__,
                "src"       : self.resource_name,
                "method"    : "DbMgr.__exit__",
                "db_path"   : k,
                "message"   : "Closed in memory db",
                })

    def __repr__(self):
        return str(self.dbs.keys())

def insert(logger, resource_name, sql_dir, db_dir, new_files):
    chlogger = logger.getChild(__name__)
    with DbMgr(chlogger, resource_name) as dbmgr:
        new_files_count = len(new_files)
        if not os.path.exists(db_dir):
            os.makedirs(db_dir)
        if not os.path.exists(db_dir):
            raise Exception("Failed to create db_dir: %s" % db_dir)
        log.info(chlogger, {
            "name"      : __name__,
            "src"       : resource_name,
            "method"    : "insert",
            "sql_dir"   : sql_dir,
            "db_dir"    : db_dir,
            "new_files" : new_files_count,
            })
        for (idx, sql_file_name) in enumerate(new_files):
            yield insert_file(logger, resource_name, dbmgr, sql_dir, db_dir, sql_file_name, idx, depth=0, max_depth=5)
       
        save_dir        = os.path.join(os.path.dirname(db_dir), "save")
        save_state_file = os.path.join(save_dir, "state.txt")
        db_files        = filesystem.glob_dir(db_dir, ".db")
        with open(save_state_file, 'w') as f:
            for dbf in db_files:
                f.write("%s\n" % dbf)
                log.debug(logger, {
                    "name"      : __name__,
                    "method"    : "insert",
                    "resource"  : resource_name,
                    "db_file"   : dbf,
                    "state_file": save_state_file,
                    "message"   : "added db_file to state file",
                    })

def insert_file(logger, resource_name, dbmgr, sql_dir, db_dir, sql_file_name, idx, depth, max_depth):
    chlogger    = logger.getChild(__name__)
    db_name     = gen_db_name(resource_name, depth)
    sql_file    = os.path.join(sql_dir, sql_file_name)
    db_file     = os.path.join(db_dir, db_name)
    if depth > max_depth:
        log.error(chlogger, {
            "name"      : __name__,
            "src"       : resource_name,
            "method"    : "insert_file",
            "db_file"   : db_file,
            "file_idx"  : idx,
            "sql_file"  : sql_file,
            "depth"     : depth,
            "max_depth" : max_depth,
            "dbmgr"     : str(dbmgr),
            "ERROR"     :"insert sql_file failed, max_depth exceeded",
            })
        return

    log.info(chlogger, {
        "name"      : __name__,
        "src"       : resource_name,
        "method"    : "insert_file",
        "db_file"   : db_file,
        "file_idx"  : idx,
        "sql_file"  : sql_file,
        "depth"     : depth,
        "dbmgr"     : str(dbmgr),
        "message"   : "started",
        })
        
    cnx = dbmgr.get(db_file)
    try:
        with open(sql_file, 'r') as sf:
            log.debug(chlogger, {
                "name"      : __name__,
                "src"       : resource_name,
                "method"    : "insert_file",
                "db_file"   : db_file,
                "file_idx"  : idx,
                "sql_file"  : sql_file,
                "depth"     : depth,
                "dbmgr"     : str(dbmgr),
                "message"   : "started",
                })
            cnx.executescript(sf.read())
            log.debug(chlogger, {
                "name"      : __name__,
                "src"       : resource_name,
                "method"    : "insert_file",
                "db_file"   : db_file,
                "file_idx"  : idx,
                "sql_file"  : sql_file,
                "depth"     : depth,
                "dbmgr"     : str(dbmgr),
                "message"   : "completed",
                })
        return sql_file_name
    except Exception as e:
        log.error(chlogger, {
            "name"      : __name__,
            "src"       : resource_name,
            "method"    : "insert_file",
            "file_idx"  : idx,
            "db_file"   : db_file,
            "sql_file"  : sql_file,
            "depth"     : depth,
            "dbmgr"     : str(dbmgr),
            "ERROR"     : "insert sql_file failed",
            "exception": str(e),
            })
        insert_file(logger, resource_name, dbmgr, sql_dir, db_dir, sql_file, idx, depth+1, max_depth)

def gen_db_name(resource_name, depth):
    return "%s_%02d.db" % (resource_name, depth)



# edl : common library for the energy-dashboard tool-chain
# Copyright (C) 2019  Todd Greenwood-Geer (Enviro Software Solutions, LLC)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
web.py : download resources from a URL
"""

from edl.resources import filesystem
from edl.resources import log
from stat import S_IREAD, S_IRGRP, S_IROTH
import logging
import os
import pdb
import requests
import time
import traceback

def generate_urls(logger, date_pairs, url_template, date_format="%Y%m%d"):
    """
    Generate download urls for the provided date_pairs.

    date_pairs      : list of tuples with (start, end) dates
    url_template    : contains a _START_ and _END_ strings which will be replaced
                      by (start,end) tuples formated by the date_format
    date_format     : format string for the start and end dates

    TODO/BIKESHED   : replace _X_ with Jinja mustache templates {{}}
    """
    chlogger = logger.getChild(__name__)
    for (start, end) in date_pairs:
        s = start.strftime(date_format)
        e = end.strftime(date_format)
        url = url_template.replace("_START_", s).replace("_END_", e)
        log.debug(chlogger, {
            "name"      : __name__,
            "method"    : "generate_urls",
            "start"     : s,
            "end"       : e,
            "url"       : url
            })
        yield url



def download(logger, resource_name, delay, urls, state_file, path):
    """
    urls        : list of urls to download
    state_file  : list of urls that have already been downloaded
    path        : path to write downloaded files to
    """
    chlogger = logger.getChild(__name__)
    downloaded = []
    prev_downloaded = set()
    if os.path.exists(state_file):
        with open(state_file, "r") as f:
            prev_downloaded = set([line.rstrip() for line in f])

    status = {'manifest': 0, 'filesystem': 0, 'downloaded': 0, 'error': 0}

    for url in urls:
        try:
            filename = filesystem.url2filename(url)
            if url in prev_downloaded:
                log.debug(chlogger, {"src":resource_name, "action":'skip_download', "url":url, "file":filename, "msg":'url exists in download manifest'})
                status['manifest'] += 1
                continue
            target_file = os.path.join(path, filename)
            if os.path.exists(target_file):
                log.debug(chlogger, {"src":resource_name, "action":'skip_download', "url":url, "file":filename, "msg":'file exists locally, updating manifest'})
                # update the state_file with files that were found on disk
                downloaded.append(url)
                status['filesystem'] += 1
                continue
            r = requests.get(url)
            if r.status_code == 200:
                with open(target_file, 'wb') as fd:
                    for chunk in r.iter_content(chunk_size=128):
                        fd.write(chunk)
                downloaded.append(url)
                status['downloaded'] += 1
                log.debug(chlogger, {"src":resource_name, "action":'download', "url":url, "file":filename})
            else:
                log.error(chlogger, {"src":resource_name, "action":'download', "url":url, "file":filename, "status_code":r.status_code, "ERROR":'http_request_failed'})
        except Exception as e:
            log.error(chlogger, {"src":resource_name, "action":'download', "url":url, "ERROR": "http_request_failed", "exception" : str(e), "traceback": str(tb = traceback.format_exc())})
            status['error'] += 1
        # TODO: this is such a hack
        time.sleep(delay)
        # ensure that all files in the download directery are read only
        for f in filesystem.glob_dir(path, ".zip"):
            os.chmod(os.path.join(path, f), S_IREAD|S_IRGRP|S_IROTH)
        log.info(chlogger, {                                        \
                "src"                   : resource_name,            \
                "action"                : 'download',               \
                "url"                   : url,                      \
                'skipped_in_manifest'   : status['manifest'],       \
                'skipped_in_filesystem' : status['filesystem'],     \
                'downloaded'            : status['downloaded'],     \
                'error'                 : status['error'],          \
                })
    return downloaded



#!/usr/bin/env python
"""
Makes working with XML feel like you are working with JSON

Copyright (C) 2012 Martin Blech and individual contributors.

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Notes: this is just hacked into my this project and should be removed
once this PR is accepted:

    https://github.com/martinblech/xmltodict/pull/224

--Todd Greenwood-Geer

"""

try:
    from defusedexpat import pyexpat as expat
except ImportError:
    from xml.parsers import expat
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesImpl
try:  # pragma no cover
    from cStringIO import StringIO
except ImportError:  # pragma no cover
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

from collections import OrderedDict
from inspect import isgenerator

try:  # pragma no cover
    _basestring = basestring
except NameError:  # pragma no cover
    _basestring = str
try:  # pragma no cover
    _unicode = unicode
except NameError:  # pragma no cover
    _unicode = str

__author__ = 'Martin Blech'
__version__ = '0.12.0'
__license__ = 'MIT'


class ParsingInterrupted(Exception):
    pass


class _DictSAXHandler(object):
    def __init__(self,
                 item_depth=0,
                 item_callback=lambda *args: True,
                 xml_attribs=True,
                 attr_prefix='@',
                 cdata_key='#text',
                 force_cdata=False,
                 cdata_separator='',
                 postprocessor=None,
                 dict_constructor=OrderedDict,
                 strip_whitespace=True,
                 namespace_separator=':',
                 namespaces=None,
                 force_list=None,
                 comment_key='#comment',
                 strip_namespaces=False
                 ):
        self.path = []
        self.stack = []
        self.data = []
        self.item = None
        self.item_depth = item_depth
        self.xml_attribs = xml_attribs
        self.item_callback = item_callback
        self.attr_prefix = attr_prefix
        self.cdata_key = cdata_key
        self.force_cdata = force_cdata
        self.cdata_separator = cdata_separator
        self.postprocessor = postprocessor
        self.dict_constructor = dict_constructor
        self.strip_whitespace = strip_whitespace
        self.namespace_separator = namespace_separator
        self.namespaces = namespaces
        self.namespace_declarations = OrderedDict()
        self.force_list = force_list
        self.comment_key = comment_key
        self.strip_namespaces = strip_namespaces

    def _build_name(self, full_name):
        if self.namespaces is None and not self.strip_namespaces:
            return full_name
        i = full_name.rfind(self.namespace_separator)
        if i == -1:
            return full_name
        namespace, name = full_name[:i], full_name[i+1:]
        if self.strip_namespaces:
            return name
        try:
            short_namespace = self.namespaces[namespace]
        except KeyError:
            short_namespace = namespace
        if not short_namespace:
            return name
        else:
            return self.namespace_separator.join((short_namespace, name))

    def _attrs_to_dict(self, attrs):
        if isinstance(attrs, dict):
            return attrs
        return self.dict_constructor(zip(attrs[0::2], attrs[1::2]))

    def startNamespaceDecl(self, prefix, uri):
        self.namespace_declarations[prefix or ''] = uri

    def startElement(self, full_name, attrs):
        name = self._build_name(full_name)
        attrs = self._attrs_to_dict(attrs)
        if attrs and self.namespace_declarations:
            attrs['xmlns'] = self.namespace_declarations
            self.namespace_declarations = OrderedDict()
        self.path.append((name, attrs or None))
        if len(self.path) > self.item_depth:
            self.stack.append((self.item, self.data))
            if self.xml_attribs:
                attr_entries = []
                for key, value in attrs.items():
                    key = self.attr_prefix+self._build_name(key)
                    if self.postprocessor:
                        entry = self.postprocessor(self.path, key, value)
                    else:
                        entry = (key, value)
                    if entry:
                        attr_entries.append(entry)
                attrs = self.dict_constructor(attr_entries)
            else:
                attrs = None
            self.item = attrs or None
            self.data = []

    def endElement(self, full_name):
        name = self._build_name(full_name)
        if len(self.path) == self.item_depth:
            item = self.item
            if item is None:
                item = (None if not self.data
                        else self.cdata_separator.join(self.data))

            should_continue = self.item_callback(self.path, item)
            if not should_continue:
                raise ParsingInterrupted()
        if len(self.stack):
            data = (None if not self.data
                    else self.cdata_separator.join(self.data))
            item = self.item
            self.item, self.data = self.stack.pop()
            if self.strip_whitespace and data:
                data = data.strip() or None
            if data and self.force_cdata and item is None:
                item = self.dict_constructor()
            if item is not None:
                if data:
                    self.push_data(item, self.cdata_key, data)
                self.item = self.push_data(self.item, name, item)
            else:
                self.item = self.push_data(self.item, name, data)
        else:
            self.item = None
            self.data = []
        self.path.pop()

    def characters(self, data):
        if not self.data:
            self.data = [data]
        else:
            self.data.append(data)

    def comments(self, data):
        if self.strip_whitespace:
            data = data.strip()
        self.item = self.push_data(self.item, self.comment_key, data)

    def push_data(self, item, key, data):
        if self.postprocessor is not None:
            result = self.postprocessor(self.path, key, data)
            if result is None:
                return item
            key, data = result
        if item is None:
            item = self.dict_constructor()
        try:
            value = item[key]
            if isinstance(value, list):
                value.append(data)
            else:
                item[key] = [value, data]
        except KeyError:
            if self._should_force_list(key, data):
                item[key] = [data]
            else:
                item[key] = data
        return item

    def _should_force_list(self, key, value):
        if not self.force_list:
            return False
        if isinstance(self.force_list, bool):
            return self.force_list
        try:
            return key in self.force_list
        except TypeError:
            return self.force_list(self.path[:-1], key, value)


def parse(xml_input, encoding=None, expat=expat, process_namespaces=False,
          namespace_separator=':', disable_entities=True, process_comments=False, **kwargs):
    """Parse the given XML input and convert it into a dictionary.

    `xml_input` can either be a `string`, a file-like object, or a generator of strings.

    If `xml_attribs` is `True`, element attributes are put in the dictionary
    among regular child elements, using `@` as a prefix to avoid collisions. If
    set to `False`, they are just ignored.

    Simple example::

        >>> import xmltodict
        >>> doc = xmltodict.parse(\"\"\"
        ... <a prop="x">
        ...   <b>1</b>
        ...   <b>2</b>
        ... </a>
        ... \"\"\")
        >>> doc['a']['@prop']
        u'x'
        >>> doc['a']['b']
        [u'1', u'2']

    If `item_depth` is `0`, the function returns a dictionary for the root
    element (default behavior). Otherwise, it calls `item_callback` every time
    an item at the specified depth is found and returns `None` in the end
    (streaming mode).

    The callback function receives two parameters: the `path` from the document
    root to the item (name-attribs pairs), and the `item` (dict). If the
    callback's return value is false-ish, parsing will be stopped with the
    :class:`ParsingInterrupted` exception.

    Streaming example::

        >>> def handle(path, item):
        ...     print('path:%s item:%s' % (path, item))
        ...     return True
        ...
        >>> xmltodict.parse(\"\"\"
        ... <a prop="x">
        ...   <b>1</b>
        ...   <b>2</b>
        ... </a>\"\"\", item_depth=2, item_callback=handle)
        path:[(u'a', {u'prop': u'x'}), (u'b', None)] item:1
        path:[(u'a', {u'prop': u'x'}), (u'b', None)] item:2

    The optional argument `postprocessor` is a function that takes `path`,
    `key` and `value` as positional arguments and returns a new `(key, value)`
    pair where both `key` and `value` may have changed. Usage example::

        >>> def postprocessor(path, key, value):
        ...     try:
        ...         return key + ':int', int(value)
        ...     except (ValueError, TypeError):
        ...         return key, value
        >>> xmltodict.parse('<a><b>1</b><b>2</b><b>x</b></a>',
        ...                 postprocessor=postprocessor)
        OrderedDict([(u'a', OrderedDict([(u'b:int', [1, 2]), (u'b', u'x')]))])

    You can pass an alternate version of `expat` (such as `defusedexpat`) by
    using the `expat` parameter. E.g:

        >>> import defusedexpat
        >>> xmltodict.parse('<a>hello</a>', expat=defusedexpat.pyexpat)
        OrderedDict([(u'a', u'hello')])

    You can use the force_list argument to force lists to be created even
    when there is only a single child of a given level of hierarchy. The
    force_list argument is a tuple of keys. If the key for a given level
    of hierarchy is in the force_list argument, that level of hierarchy
    will have a list as a child (even if there is only one sub-element).
    The index_keys operation takes precendence over this. This is applied
    after any user-supplied postprocessor has already run.

        For example, given this input:
        <servers>
          <server>
            <name>host1</name>
            <os>Linux</os>
            <interfaces>
              <interface>
                <name>em0</name>
                <ip_address>10.0.0.1</ip_address>
              </interface>
            </interfaces>
          </server>
        </servers>

        If called with force_list=('interface',), it will produce
        this dictionary:
        {'servers':
          {'server':
            {'name': 'host1',
             'os': 'Linux'},
             'interfaces':
              {'interface':
                [ {'name': 'em0', 'ip_address': '10.0.0.1' } ] } } }

        `force_list` can also be a callable that receives `path`, `key` and
        `value`. This is helpful in cases where the logic that decides whether
        a list should be forced is more complex.


        If `process_comment` is `True` then comment will be added with comment_key
        (default=`'#comment'`) to then tag which contains comment

            For example, given this input:
            <a>
              <b>
                <!-- b comment -->
                <c>
                    <!-- c comment -->
                    1
                </c>
                <d>2</d>
              </b>
            </a>

            If called with process_comment=True, it will produce
            this dictionary:
            'a': {
                'b': {
                    '#comment': 'b comment',
                    'c': {

                        '#comment': 'c comment',
                        '#text': '1',
                    },
                    'd': '2',
                },
            }
    """
    handler = _DictSAXHandler(namespace_separator=namespace_separator,
                              **kwargs)
    if isinstance(xml_input, _unicode):
        if not encoding:
            encoding = 'utf-8'
        xml_input = xml_input.encode(encoding)
    if not process_namespaces:
        namespace_separator = None
    parser = expat.ParserCreate(
        encoding,
        namespace_separator
    )
    try:
        parser.ordered_attributes = True
    except AttributeError:
        # Jython's expat does not support ordered_attributes
        pass
    parser.StartNamespaceDeclHandler = handler.startNamespaceDecl
    parser.StartElementHandler = handler.startElement
    parser.EndElementHandler = handler.endElement
    parser.CharacterDataHandler = handler.characters
    if process_comments:
        parser.CommentHandler = handler.comments
    parser.buffer_text = True
    if disable_entities:
        try:
            # Attempt to disable DTD in Jython's expat parser (Xerces-J).
            feature = "http://apache.org/xml/features/disallow-doctype-decl"
            parser._reader.setFeature(feature, True)
        except AttributeError:
            # For CPython / expat parser.
            # Anything not handled ends up here and entities aren't expanded.
            parser.DefaultHandler = lambda x: None
            # Expects an integer return; zero means failure -> expat.ExpatError.
            parser.ExternalEntityRefHandler = lambda *x: 1
    if hasattr(xml_input, 'read'):
        parser.ParseFile(xml_input)
    elif isgenerator(xml_input):
        for chunk in xml_input:
            parser.Parse(chunk,False)
        parser.Parse(b'',True)
    else:
        parser.Parse(xml_input, True)
    return handler.item


def _process_namespace(name, namespaces, ns_sep=':', attr_prefix='@'):
    if not namespaces:
        return name
    try:
        ns, name = name.rsplit(ns_sep, 1)
    except ValueError:
        pass
    else:
        ns_res = namespaces.get(ns.strip(attr_prefix))
        name = '{}{}{}{}'.format(
            attr_prefix if ns.startswith(attr_prefix) else '',
            ns_res, ns_sep, name) if ns_res else name
    return name


def _emit(key, value, content_handler,
          attr_prefix='@',
          cdata_key='#text',
          depth=0,
          preprocessor=None,
          pretty=False,
          newl='\n',
          indent='\t',
          namespace_separator=':',
          namespaces=None,
          full_document=True):
    key = _process_namespace(key, namespaces, namespace_separator, attr_prefix)
    if preprocessor is not None:
        result = preprocessor(key, value)
        if result is None:
            return
        key, value = result
    if (not hasattr(value, '__iter__')
            or isinstance(value, _basestring)
            or isinstance(value, dict)):
        value = [value]
    for index, v in enumerate(value):
        if full_document and depth == 0 and index > 0:
            raise ValueError('document with multiple roots')
        if v is None:
            v = OrderedDict()
        elif isinstance(v, bool):
            if v:
                v = _unicode('true')
            else:
                v = _unicode('false')
        elif not isinstance(v, dict):
            v = _unicode(v)
        if isinstance(v, _basestring):
            v = OrderedDict(((cdata_key, v),))
        cdata = None
        attrs = OrderedDict()
        children = []
        for ik, iv in v.items():
            if ik == cdata_key:
                cdata = iv
                continue
            if ik.startswith(attr_prefix):
                ik = _process_namespace(ik, namespaces, namespace_separator,
                                        attr_prefix)
                if ik == '@xmlns' and isinstance(iv, dict):
                    for k, v in iv.items():
                        attr = 'xmlns{}'.format(':{}'.format(k) if k else '')
                        attrs[attr] = _unicode(v)
                    continue
                if not isinstance(iv, _unicode):
                    iv = _unicode(iv)
                attrs[ik[len(attr_prefix):]] = iv
                continue
            children.append((ik, iv))
        if pretty:
            content_handler.ignorableWhitespace(depth * indent)
        content_handler.startElement(key, AttributesImpl(attrs))
        if pretty and children:
            content_handler.ignorableWhitespace(newl)
        for child_key, child_value in children:
            _emit(child_key, child_value, content_handler,
                  attr_prefix, cdata_key, depth+1, preprocessor,
                  pretty, newl, indent, namespaces=namespaces,
                  namespace_separator=namespace_separator)
        if cdata is not None:
            content_handler.characters(cdata)
        if pretty and children:
            content_handler.ignorableWhitespace(depth * indent)
        content_handler.endElement(key)
        if pretty and depth:
            content_handler.ignorableWhitespace(newl)


def unparse(input_dict, output=None, encoding='utf-8', full_document=True,
            short_empty_elements=False,
            **kwargs):
    """Emit an XML document for the given `input_dict` (reverse of `parse`).

    The resulting XML document is returned as a string, but if `output` (a
    file-like object) is specified, it is written there instead.

    Dictionary keys prefixed with `attr_prefix` (default=`'@'`) are interpreted
    as XML node attributes, whereas keys equal to `cdata_key`
    (default=`'#text'`) are treated as character data.

    The `pretty` parameter (default=`False`) enables pretty-printing. In this
    mode, lines are terminated with `'\n'` and indented with `'\t'`, but this
    can be customized with the `newl` and `indent` parameters.

    """
    if full_document and len(input_dict) != 1:
        raise ValueError('Document must have exactly one root.')
    must_return = False
    if output is None:
        output = StringIO()
        must_return = True
    if short_empty_elements:
        content_handler = XMLGenerator(output, encoding, True)
    else:
        content_handler = XMLGenerator(output, encoding)
    if full_document:
        content_handler.startDocument()
    for key, value in input_dict.items():
        _emit(key, value, content_handler, full_document=full_document,
              **kwargs)
    if full_document:
        content_handler.endDocument()
    if must_return:
        value = output.getvalue()
        try:  # pragma no cover
            value = value.decode(encoding)
        except AttributeError:  # pragma no cover
            pass
        return value


if __name__ == '__main__':  # pragma: no cover
    import sys
    import marshal
    try:
        stdin = sys.stdin.buffer
        stdout = sys.stdout.buffer
    except AttributeError:
        stdin = sys.stdin
        stdout = sys.stdout

    (item_depth,) = sys.argv[1:]
    item_depth = int(item_depth)

    def handle_item(path, item):
        marshal.dump((path, item), stdout)
        return True

    try:
        root = parse(stdin,
                     item_depth=item_depth,
                     item_callback=handle_item,
                     dict_constructor=dict)
        if item_depth == 0:
            handle_item([], root)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python3
# edl : common library for the energy-dashboard tool-chain
# Copyright (C) 2019  Todd Greenwood-Geer (Enviro Software Solutions, LLC)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import base64
import traceback
from enum import Enum
import codecs
import datetime as dt
import fileinput
import json
import logging
import os
import pdb
import pprint
import re
import sys
import uuid
#import xmltodict
from edl.resources import log
from edl.external import xmltodict
import sqlite3

class Walker(object):
    def __init__(self, logger, dict_handler_func=None, list_handler_func=None, item_handler_func=None):
        def default_dict_handler_func(stack):
            pass
        def default_list_handler_func(stack):
            pass
        def default_item_handler_func(stack):
            pass
        self.logger = logger
        self.dict_handler_func = dict_handler_func or default_dict_handler_func
        self.list_handler_func = list_handler_func or default_list_handler_func
        self.item_handler_func = item_handler_func or default_item_handler_func

    def walk(self, name, obj):
        self._walk([(name, obj)])

    def _walk(self, stack):
        """
        Walk object tree and inkoke handlers based on object type (dict, list, or item).
        """
        (name, obj) = stack[-1]
        if isinstance(obj, dict):
            self.dict_handler_func(stack)
            for k, v in obj.items():
                self._walk(stack + [(k, v)])
        elif isinstance(obj, list):
            self.list_handler_func(stack)
            for idx, item in enumerate(obj):
                self._walk(stack + [(idx, item)])
        else:
            self.item_handler_func(stack)
        stack.pop()

class SqlTypeEnum(Enum):
    """
    Convert datatypes found in XML into Sqlite3 datatypes:

        NULL. The value is a NULL value.
        INTEGER. The value is a signed integer, stored in 1, 2, 3, 4, 6, or 8 bytes depending on the magnitude of the value.
        REAL. The value is a floating point value, stored as an 8-byte IEEE floating point number.
        TEXT. The value is a text string, stored using the database encoding (UTF-8, UTF-16BE or UTF-16LE).
        BLOB. The value is a blob of data, stored exactly as it was input.

    See : https://www.sqlite.org/datatype3.html
    """
    NULL        = 1
    INTEGER     = 2
    REAL        = 3
    TEXT        = 4
    BLOB        = 5

    def type_of(element):
        """
        Return the SqlType for the element.
        """
        if element is None:
            return SqlTypeEnum.NULL
        try:
            int(element)
            return SqlTypeEnum.INTEGER
        except:
            try:
                float(element)
                return SqlTypeEnum.REAL
            except:
                try:
                    str(element)
                    return SqlTypeEnum.TEXT
                except:
                    return SqlTypeEnum.BLOB

def sql_type_str(e):
    """
    Ideally this would hang of the enum class above, but making it __repr__ didn't work, strangely.
    TODO: 
    """
    if e == SqlTypeEnum.NULL: return "NULL"
    if e == SqlTypeEnum.INTEGER: return "INTEGER"
    if e == SqlTypeEnum.REAL: return "REAL"
    if e == SqlTypeEnum.TEXT: return "TEXT"
    if e == SqlTypeEnum.BLOB: return "BLOB"

class Table(object):
    def __init__(self, name, parent):
        self.name = name
        """table name"""

        self.parent = parent
        """name parent of this table, e.g. where this table has foreign key references to"""

        self.columns = set(['id'])
        """columns in this table"""

        self.last_id = None
        """dirty hack to track last id of a list"""

    def primary_key(self):
        """primary key is always ['id']"""
        assert 'id' in self.columns
        return ['id']

    def __repr__(self):
        return "name: %s, columns: %s, parent: %s, " % (self.name, self.columns, self.parent)

class XML2SQLTransormer():
    """
    OASIS Reports have the following nested structure:

    Example:

        {
          "OASISReport": {
            "@xmlns": "http://www.caiso.com/soa/OASISReport_v1.xsd",
            "MessageHeader": {
              "TimeDate": "2019-08-09T15:27:07-00:00",
              "Source": "OASIS",
              "Version": "v20131201"
            },
            "MessagePayload": {
              "RTO": {
                "name": "CAISO",
                "REPORT_ITEM": [
                  {
                    "REPORT_HEADER": {
                      "SYSTEM": "OASIS",
                      "TZ": "PPT",
                      "REPORT": "AS_MILEAGE_CALC",
                      "UOM": "MW",
                      "INTERVAL": "ENDING",
                      "SEC_PER_INTERVAL": "3600"
                    },
                    "REPORT_DATA": {
                      "DATA_ITEM": "RMD_AVG_MIL",
                      "RESOURCE_NAME": "AS_CAISO_EXP",
                      "OPR_DATE": "2018-02-24",
                      "INTERVAL_NUM": "24",
                      "INTERVAL_START_GMT": "2018-02-25T07:00:00-00:00",
                      "INTERVAL_END_GMT": "2018-02-25T08:00:00-00:00",
                      "VALUE": "2426.9"
                    }
                  },
            ...
        }

    Example transforms:

    # 1. Dict A maps to a list of anonymous dicts -> named dicts -> items, e.g. {A:[{T:{a:1,b:2,}},]}
    # 1. Dict A maps to a list of anonymous dicts -> items, e.g. {A:[{a:1,b:2,},]}
             => {A:[{T:{a:1,b:2,}},]}

   


    Example2:
    {
    "OASISReport": {
        "MessageHeader": {
            "Source": "OASIS",
            "TimeDate": "2019-09-07T14:50:16-00:00",
            "Version": "v20131201"
        },
        "MessagePayload": {
            "RTO": {
                "DISCLAIMER_ITEM": {
                    "DISCLAIMER": "The contents of these pages are subject to change without notice.  Decisions based on information contained within the California ISO's web site are the visitor's sole responsibility."
                },
                "REPORT_ITEM": [
                    {
                        "REPORT_DATA": [
                            {
                                "DATA_ITEM": "RMD_AVG_MIL",
                                "INTERVAL_END_GMT": "2019-09-05T08:00:00-00:00",
                                "INTERVAL_NUM": "1",
                                "INTERVAL_START_GMT": "2019-09-05T07:00:00-00:00",
                                "OPR_DATE": "2019-09-05",
                                "RESOURCE_NAME": "AS_CAISO_EXP",
                                "VALUE": "1750.69"
                            },

odict_keys(['DATA_ITEM', 'RESOURCE_NAME', 'OPR_DATE', 'INTERVAL_NUM', 'INTERVAL_START_GMT', 'INTERVAL_END_GMT', 'VALUE'])


                            {
                                "DATA_ITEM": "RMD_AVG_MIL",
                                "INTERVAL_END_GMT": "2019-09-05T09:00:00-00:00",
                                "INTERVAL_NUM": "2",
                                "INTERVAL_START_GMT": "2019-09-05T08:00:00-00:00",
                                "OPR_DATE": "2019-09-05",
                                "RESOURCE_NAME": "AS_CAISO_EXP",
                                "VALUE": "1506.88"
                            },


    Example:

        Running this tooling on an XML file generates the DDL create table statements necessary:

        $ ./energy-dashboard-lib/edl/resources/xmlparser.py 20180224_20180225_AS_MILEAGE_CALC_N_20190809_08_27_07_v1.xml

        CREATE TABLE IF NOT EXISTS oasisreport (xmlns TEXT, PRIMARY KEY (xmlns));
        CREATE TABLE IF NOT EXISTS messageheader (timedate TEXT, source TEXT, version TEXT, oasisreport_xmlns TEXT, FOREIGN KEY (oasisreport_xmlns) REFERENCES oasisreport(xmlns), PRIMARY KEY (timedate, source, version));
        CREATE TABLE IF NOT EXISTS messagepayload (id INTEGER, oasisreport_xmlns TEXT, FOREIGN KEY (oasisreport_xmlns) REFERENCES oasisreport(xmlns), PRIMARY KEY (id));
        CREATE TABLE IF NOT EXISTS rto (name TEXT, messagepayload_id INTEGER, FOREIGN KEY (messagepayload_id) REFERENCES messagepayload(id), PRIMARY KEY (name));
        CREATE TABLE IF NOT EXISTS report_item (id INTEGER, rto_name TEXT, FOREIGN KEY (rto_name) REFERENCES rto(name), PRIMARY KEY (id));
        CREATE TABLE IF NOT EXISTS report_header (system TEXT, tz TEXT, report TEXT, uom TEXT, interval TEXT, sec_per_interval INTEGER, report_item_id INTEGER, FOREIGN KEY (report_item_id) REFERENCES report_item(id), PRIMARY KEY (system, tz, report, uom, interval, sec_per_interval));
        CREATE TABLE IF NOT EXISTS report_data (data_item TEXT, resource_name TEXT, opr_date TEXT, interval_num INTEGER, interval_start_gmt TEXT, interval_end_gmt TEXT, value REAL, report_item_id INTEGER, FOREIGN KEY (report_item_id) REFERENCES report_item(id), PRIMARY KEY (data_item, resource_name, opr_date, interval_num, interval_start_gmt, interval_end_gmt));
        CREATE TABLE IF NOT EXISTS disclaimer_item (disclaimer TEXT, rto_name TEXT, FOREIGN KEY (rto_name) REFERENCES rto(name), PRIMARY KEY (disclaimer));


    Example:

        The output DDL can be quickly sanity checked by running it through sqlite3:

        $ python energy-dashboard-lib/edl/resources/xmlparser.py ./20180224_20180225_AS_MILEAGE_CALC_N_20190809_08_27_07_v1.xml | sqlite3

        No errors, so the structure is good.

    https://sqlite.org/foreignkeys.html
    https://docs.python.org/3/library/json.html
    """
    NULL_MIGRATION  = set([SqlTypeEnum.REAL, SqlTypeEnum.INTEGER, SqlTypeEnum.TEXT, SqlTypeEnum.BLOB]) 
    BLOB_MIGRATION  = set([SqlTypeEnum.REAL, SqlTypeEnum.INTEGER, SqlTypeEnum.TEXT]) 
    def __init__(self, logger, xmlfile):
        """
        Following 'Elegant Objects' style, constructor only sets vars. All
        work is delayed until needed.
        """
        self.logger             = logger
        """use log.info(self.logger, {})"""

        self.xmlfile            = xmlfile
        """xmlfile : file obect""" 

        self.json               = None
        """json : json object (typically a dict) resulting from parsing the xmlfile object"""

        self.sql_types          = {'id': SqlTypeEnum.TEXT}
        """sql_types : map : xml element name -> sqlite3 value type"""

        self.tables             = {}
        """tables : map : table name -> Table(), which are just table definitions for sql inserts"""

        self.root               = 'root'
        """root : name of the root node which is really the document root, for which no table is generated"""

    def parse(self):
        """
        Parse the loaded xmlfile and load into the self.json object.
        """
        self.json = xmltodict.parse(self.xmlfile.read(), process_namespaces=True, strip_namespaces=True)
        if self.logger.isEnabledFor(logging.DEBUG):
            print(json.dumps(self.json, indent=4, sort_keys=True))
        # allow method chaining
        return self

    def scan_all(self):
        """
        Recursive scan to build the types, tables, and table relations.
        """
        def handle_list_or_dict(stack):
            (child_name, obj, parent_name, pobj) = self.find_parent_child_tables(stack)
            if child_name == self.root:
                return
            if child_name not in self.tables:
                self.tables[child_name] = Table(name=child_name, parent=parent_name)
        def handle_item(stack):
            (item_name, obj) = stack[-1]
            if item_name not in self.sql_types:
                sqltype = SqlTypeEnum.type_of(obj)
                self.sql_types[item_name] = sqltype
            else:
                # item_name was already in sql_types, but here we need to allow some flexibilty.
                # bad formatted xml with blanks or nulls or missing data will leave the item type as NULL
                # when we know from later data that the type is actually more specific.
                # so here we allow NULL to be upgraded to something else.
                # allow NULL    -> REAL, INTEGER, TEXT, BLOB
                # allow BLOB    -> REAl, INTEGER, TEXT
                # allow INTEGER -> REAL
                existing_type = self.sql_types[item_name]
                new_type = SqlTypeEnum.type_of(obj)
                if existing_type == SqlTypeEnum.NULL and new_type in XML2SQLTransormer.NULL_MIGRATION:
                    self.sql_types[item_name] = new_type
                elif existing_type == SqlTypeEnum.BLOB and new_type in XML2SQLTransormer.BLOB_MIGRATION:
                    self.sql_types[item_name] = new_type
                elif existing_type == SqlTypeEnum.INTEGER and new_type == SqlTypeEnum.REAL:
                    self.sql_types[item_name] = new_type
            # add item to the table columns
            (child_name, obj, parent_name, pobj) = self.find_parent_child_tables(stack)
            t = self.tables[child_name]
            t.columns.add(item_name)

        Walker(self.logger, item_handler_func=handle_item, list_handler_func=handle_list_or_dict, dict_handler_func=handle_list_or_dict).walk(self.root, self.json)

        # post scan add 'id' to empty tables
        for k,v in self.tables.items():
            assert 'id' in v.columns
        # post scan to convert NULL sql_types to TEXT
        for k,v in self.sql_types.items():
            if v == SqlTypeEnum.NULL:
                self.sql_types[k] = SqlTypeEnum.TEXT
        for k,v in self.sql_types.items():
            log.debug(self.logger, {
                "name"      : __name__,
                "method"    : "scan_all",
                "column"    : str(k),
                "type"      : sql_type_str(v)
                })
        for k,v in self.tables.items():
            log.debug(self.logger, {
                "name"      : __name__,
                "method"    : "scan_all",
                "table"     : str(k),
                "def"       : str(v)
                })
        return self

    def find_table(self, stack):
        x = 1
        while True:
            if x > len(stack):
                return (None, None, None)
            (name, curobj) = stack[-x]
            if (SqlTypeEnum.type_of(name) == SqlTypeEnum.TEXT) and (isinstance(curobj, dict) or isinstance(curobj, list) and name != self.root):
                return (name, curobj, -x)
            x = x + 1

    def find_parent_child_tables(self, stack):
        (child, obj, cidx) = self.find_table(stack)
        (parent, pobj, pidx) = self.find_table(stack[:cidx].copy())
        return (child, obj, parent, pobj)

    def table2ddl(self, t):
        local_column_tuples         = [(self.sqlite_sanitize(c), sql_type_str(self.sql_types[c])) for c in t.columns]
        parent_column_pname_type    = []
        foreign_key_str_lst         = []
        if t.parent != None and t.parent in self.tables:
            parent = self.tables[t.parent]
            parent_column_pname_type = ["%s_%s %s"%(self.sqlite_sanitize(parent.name), a,b) for (a,b) in [(self.sqlite_sanitize(c), sql_type_str(self.sql_types[c])) for c in parent.primary_key()]]
            parent_primary_key_columns = ["%s_%s" % (parent.name, pk) for pk in parent.primary_key()]
            foreign_key_str_lst.append("FOREIGN KEY ({foreign_keys}) REFERENCES {parent_table}({parent_pk_cols})".format(
                parent_table=self.sqlite_sanitize(parent.name),
                foreign_keys=", ".join(self.sqlite_sanitize_all(parent_primary_key_columns)),
                parent_pk_cols=", ".join(self.sqlite_sanitize_all(parent.primary_key()))))
        all_columns = []
        all_columns.extend(local_column_tuples)
        all_columns_lst = ["%s %s"%(a,b) for (a,b) in all_columns]
        all_columns_lst.extend(parent_column_pname_type)
        all_columns_lst.extend(foreign_key_str_lst)
        return "CREATE TABLE IF NOT EXISTS {name} ({columns}, PRIMARY KEY ({primary_key}));".format(
                name=self.sqlite_sanitize(t.name), 
                columns=", ".join(all_columns_lst),
                primary_key=", ".join(self.sqlite_sanitize_all(t.primary_key())))

    def ddl(self):
        for name, table in self.tables.items():
            yield self.table2ddl(table)
    
    def query_sql(self):
        return []


    def insertion_sql(self):
        assert self.json is not None
        retval = []

        def sql(name, columns, values):
            sql_txt = """INSERT OR IGNORE INTO {table} ({columns}) VALUES ({values});""".format(table=name, columns=", ".join(columns), values=", ".join(values))
            return sql_txt
       
        def get_kv(keys:list, obj:dict):
            assert isinstance(obj, dict)
            sortedkeys = sorted(keys)
            for sk in sortedkeys:
                if sk not in obj:
                    raise Exception("ERROR: sk: '%s' not in obj keys: '%s'" % (sk, obj.keys()))
            values = [obj[k] for k in sortedkeys]
            return (sortedkeys, values)

        def handle_list(stack):
            (name, obj) = stack[-1]
            t = self.tables[name]
            if t.columns == {'id'}:
                synthetic_id = str(uuid.uuid4())
                t.last_id = synthetic_id

        def handle_dict(stack):
            (realname, obj) = stack[-1]
            if realname == self.root:
                return
            (name, _, parent_name, pobj) = self.find_parent_child_tables(stack)

            t = self.tables[name]

            local_keys          = []
            local_values        = []
            parent_pk_keys      = []
            parent_pk_values    = []

            if t.primary_key() == ['id'] and 'id' not in obj.keys():
                synthetic_id = str(uuid.uuid4())
                obj['id'] = synthetic_id
                t.last_id = synthetic_id

            if t.parent is not None and t.parent != self.root and t.parent == parent_name and parent_name in self.tables:
                parent_t = self.tables[parent_name]
                parent_pk = parent_t.primary_key()
                if isinstance(pobj, list):
                    assert parent_pk == ['id']
                    assert parent_t.last_id is not None
                    parent_pk_keys = parent_pk
                    parent_pk_values = [parent_t.last_id]
                else:
                    (parent_pk_keys, parent_pk_values) = get_kv(parent_pk, pobj)

            (local_keys, local_values) = get_kv(t.columns, obj)

            columns = []
            columns.extend(self.sqlite_sanitize_all(local_keys))
            columns.extend(["%s_%s"%(self.sqlite_sanitize(t.parent), self.sqlite_sanitize(k)) for k in parent_pk_keys])
    
            values  = []
            values.extend(self.sqlite_sanitize_values(local_keys, local_values))
            values.extend(self.sqlite_sanitize_values(parent_pk_keys, parent_pk_values))
            
            retval.append(sql(self.sqlite_sanitize(name), columns, values))

        #Walker(self.logger, dict_handler_func=handle_dict, list_handler_func=handle_list).walk(self.root, self.json)
        Walker(self.logger, dict_handler_func=handle_dict).walk(self.root, self.json)
        return retval


    def quote_identifier(self, s, errors="strict"):
        """
        https://stackoverflow.com/questions/6514274/how-do-you-escape-strings-for-sqlite-table-column-names-in-python
        """
        encodable = s.encode("utf-8", errors).decode("utf-8")

        nul_index = encodable.find("\x00")

        if nul_index >= 0:
            error = UnicodeEncodeError("NUL-terminated utf-8", encodable,
                                       nul_index, nul_index + 1, "NUL not allowed")
            error_handler = codecs.lookup_error(errors)
            replacement, _ = error_handler(error)
            encodable = encodable.replace("\x00", replacement)

        return "\"" + encodable.replace("\"", "\"\"") + "\""

    def sqlite_sanitize_values(self, columns, values):
        def _sani_by_type(t, v):
            if v == None:
                s_values.append("\"\"")
            else:
                if t == SqlTypeEnum.NULL:
                    # should not have NULL types (these are converted to TEXT type
                    # at the end of the initial scan)
                    s_values.append("\"\"")
                elif t == SqlTypeEnum.TEXT:
                    s_values.append(self.quote_identifier(v))
                elif t == SqlTypeEnum.INTEGER:
                    s_values.append(v)
                elif t == SqlTypeEnum.REAL:
                    s_values.append(v)
                elif t == SqlTypeEnum.BLOB:
                    s_values.append(base64.b64encode(v))
                else:
                    log.error(self.logger, {
                        "name"      : __name__,
                        "method"    : "sqlite_sanitize_values._sani_by_type",
                        "column"    : str(c),
                        "value"     : str(v),
                        "type"      : str(t),
                        "sql_types" : str(self.sql_types),
                        "ERROR"     : "Failed to sanitize value",
                        })

        s_values = []
        for (c,v) in zip(columns, values):
            if c in self.sql_types:
                _sani_by_type(self.sql_types[c], v)
            else:
                # could not find the type, which is weird b/c all the column types
                # shoud be accounted for. so we can assume this is a strange
                # error case...
                log.warning(self.logger, {
                    "name"      : __name__,
                    "method"    : "sqlite_sanitize_values",
                    "column"    : str(c),
                    "value"     : str(v),
                    "sql_types" : str(self.sql_types),
                    })
                _sani_by_type(SqlTypeEnum.type_of(v), v)

        return s_values

    def sqlite_sanitize_all(self, l):
        """
        Sanitize table names and field names prior for sqlite
        """
        return [self.sqlite_sanitize(x) for x in l]

    def sqlite_sanitize(self, s):
        """
        Sanitize table names and field names prior for sqlite
        """
        return re.sub('[^a-zA-Z0-9_]', '', s).lower()

    def _find_parent_table(self, table_name, table_relations, table_columns):
        """
        There's been some wonkiness in the path, where 'report_item' was stored multiple times in the
        path, resulting in an infinite loop.
        """
        if table_name == self.root:
            return None
        path = table_relations[table_name]
        parent = path.pop()
        while parent == table_name:
            parent = path.pop()
        if parent == self.root:
            return None
        return parent

def parse(logger, resource_name, input_files, input_dir, output_dir):
    failed_state        = os.path.join(output_dir, 'failed.txt')
    failed_input_files  = []

    if os.path.exists(failed_state):
        with open(failed_state, 'r') as fh:
            failed_input_files = [l.rstrip().rstrip() for l in fh]

    s_failed_input_files= set(failed_input_files) 
    s_input_files       = set(input_files)
    s_unprocessed_files = s_input_files - s_failed_input_files
    unprocessed_files   = sorted(list(s_unprocessed_files))

    with open(failed_state, 'a') as fh:
        for f in unprocessed_files:
            try:
                yield parse_file(logger, resource_name, f, input_dir, output_dir)
            except Exception as e:
                fh.write("%s\n" % f)
                tb = traceback.format_exc()
                log.error(logger, {
                    "src"       : resource_name, 
                    "action"    : "parse",
                    "xml_file"  : f,
                    "msg"       : "parse failed",
                    "ERROR"     : "Failed to parse xml file",
                    "exception" : str(e),
                    "trace"     : str(tb),
                    })

def parse_file(logger, resource_name, xml_input_file_name, input_dir, output_dir):
    chlogger = logger.getChild(__name__)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    (base, ext) = os.path.splitext(xml_input_file_name)
    outfile = os.path.join(output_dir, "%s.sql" % (base))
    infile  = os.path.join(input_dir, xml_input_file_name)
    total_ddl = 0
    total_sql = 0
    with open(outfile, 'w') as outfh:
        with open(infile, 'r') as infh:
            # all the work happens here
            xst = XML2SQLTransormer(chlogger, infh).parse().scan_all()
            # check that the ddl and sql is correct
            # if this fails then it means the ddl/sql combination is incorrect
            sqllst = []
            for ddl in xst.ddl():
                sqllst.append(ddl)
            for sql in xst.insertion_sql():
                sqllst.append(sql)
            sqltext = "\n".join(sqllst)
            #db = sqlite3.connect("file::memory:?cache=shared")
            db = sqlite3.connect(":memory:")
            db.executescript(sqltext)
            # all good
            outfh.write(sqltext)
    log.info(chlogger, {
        "src":resource_name, 
        "action":"parse_file",
        "infile": infile,
        "outfile":outfile,
        })
    return xml_input_file_name

if __name__ == "__main__":
    infile = sys.argv[1]
    if len(sys.argv) > 2:
        loglevel = sys.argv[2]
    else:
        loglevel = "INFO"
    log.configure_logging()
    logger = logging.getLogger(__name__)
    logger.setLevel(loglevel)
    log.debug(logger, {
        "name"      : __name__,
        "method"    : "main",
        "src"       : "xmlparser.py"
        })
    with open(infile, 'r') as f:
        xst = XML2SQLTransormer(logger, f).parse().scan_all()
        for d in xst.ddl():
            print(d)
        for sql in xst.insertion_sql():
            print(sql)
#        for query_sql in xst.query_sql():
#            print(query_sql)



# edl : common library for the energy-dashboard tool-chain
# Copyright (C) 2019  Todd Greenwood-Geer (Enviro Software Solutions, LLC)
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from edl.resources.exec import runyield
from jinja2 import Environment, PackageLoader, select_autoescape
from pathlib import Path
from shutil import make_archive, rmtree
import edl.resources.log as log
import edl.resources.filesystem as filesystem
import os
import shutil
import stat
import sys
import tarfile
import re
import requests
import json
import traceback

STAGES  = ['download', 'unzip', 'parse', 'insert', 'save', 'dist', 'arch']
DIRS    = ['zip', 'xml', 'sql', 'db', 'save', 'dist']
PROCS   = ['10_down.py', '20_unzp.py', '30_pars.py', '40_inse.py', '50_save.py', '60_dist.sh', '70_arch.py']
STAGE_DIRS = dict(zip(STAGES, DIRS))
STAGE_PROCS = dict(zip(STAGES, PROCS))


def create(logger, ed_path, feed, maintainer, company, email, url, start_date, delay):
    """
    start_date : list of numbers : [2019,09,1]
    """
    chlogger = logger.getChild(__name__)
    new_feed_dir = os.path.join(ed_path, 'data', feed)
    try:
        os.mkdir(new_feed_dir)
        log.debug(chlogger, {
            "name"      : __name__,
            "method"    : "create",
            "path"      : ed_path,
            "feed"      : feed,
            "dir"       : new_feed_dir,
            "message"   : "created directory"
            })
        template_files = [
                "LICENSE","Makefile","README.md",
                "src/10_down.py","src/20_unzp.py","src/30_pars.py",
                "src/40_inse.py", "src/50_save.py", "src/60_dist.sh", 
                "src/70_arch.py",
                "manifest.json"
                ]
        env = Environment(
            loader=PackageLoader('edl', 'templates'),
            autoescape=select_autoescape(['py'])
        )
        m = {
                'NAME'      : feed,
                'MAINTAINER': maintainer,
                'COMPANY'   : company,
                'EMAIL'     : email,
                'DATA_URL'  : url,
                'REPO_URL'  : "https://github.com/energy-analytics-project/%s" % feed,
                'START'     : start_date,
                'DELAY'     : delay
        }
        for tf in template_files:
            template    = env.get_template(tf)
            target      = os.path.join(new_feed_dir, tf)
            path        = os.path.dirname(target)
            if not os.path.exists(path):
                os.makedirs(path)
            with open(target, 'w') as f:
                f.write(template.render(m))
                log.debug(chlogger, {
                    "name"      : __name__,
                    "method"    : "create",
                    "path"      : ed_path,
                    "feed"      : feed,
                    "target"    : target,
                    "message"   : "rendered target"
                    })

        hidden_files = ['gitignore']
        for hf in hidden_files:
            template    = env.get_template(hf)
            target      = os.path.join(new_feed_dir, ".%s" % hf)
            with open(target, 'w') as f:
                f.write(template.render(m))
                log.debug(chlogger, {
                    "name"      : __name__,
                    "method"    : "create",
                    "path"      : ed_path,
                    "feed"      : feed,
                    "target"    : target,
                    "message"   : "rendered target"
                    })
        for src_file in os.listdir(os.path.join(new_feed_dir, 'src')):
            fp = os.path.join(new_feed_dir, 'src', src_file)
            f = Path(fp)
            f.chmod(f.stat().st_mode | stat.S_IEXEC)
            log.debug(chlogger, {
                "name"      : __name__,
                "method"    : "create",
                "path"      : ed_path,
                "feed"      : feed,
                "file"      : fp,
                "message"   : "chmod +x"
                })
        
        for d in DIRS:
            os.makedirs(os.path.join(new_feed_dir, d))
        return feed
    except Exception as e:
        tb = traceback.format_exc()
        log.critical(chlogger, {
            "name"      : __name__,
            "method"    : "create",
            "path"      : ed_path,
            "feed"      : feed,
            "ERROR"     : "FAILED to create feed",
            "exception" : str(e),
            "trace":str(tb),
            })
        # return nothing

def invoke(logger, feed, ed_path, command):
    chlogger = logger.getChild(__name__)
    target_dir = os.path.join(ed_path, 'data', feed)
    log.debug(chlogger, {
                "name"      : __name__,
                "method"    : "invoke",
                "path"      : ed_path,
                "feed"      : feed,
                "command"   : command
        })
    if not os.path.exists(target_dir):
        log.critical(chlogger, {
                    "name"      : __name__,
                    "method"    : "invoke",
                    "path"      : ed_path,
                    "feed"      : feed,
                    "command"   : command,
                    "target_dir": target_dir,
                    "ERROR"     : "target_dir does not exist"
            })
    else:
        return runyield([command], target_dir)

def status(logger, feed, ed_path, separator, header):
    chlogger = logger.getChild(__name__)
    target_dir = os.path.join(ed_path, 'data', feed)
    if not os.path.exists(target_dir):
        log.critical(chlogger, {
                    "name"      : __name__,
                    "method"    : "status",
                    "path"      : ed_path,
                    "feed"      : feed,
                    "separator" : separator,
                    "header"    : header,
                    "target_dir": target_dir,
                    "ERROR"     : "target_dir does not exist"
            })
    if header:
        yield separator.join(["feed name","downloaded","unzipped","parsed", "inserted", "databases"])
    txtfiles = ["zip/state.txt", "xml/state.txt", "sql/state.txt", "db/state.txt", "save/state.txt"]
    counts = [str(lines_in_file(os.path.join(target_dir, f))) for f in txtfiles]
    status = [feed]
    status.extend(counts)
    yield separator.join(status)

def pre_prune(logger, feed, ed_path, stage):
    return os.path.join(ed_path, 'data', feed, STAGE_DIRS[stage])

def prune(logger, feed, ed_path, stage):
    chlogger    = logger.getChild(__name__)
    path        = pre_prune(logger, feed, ed_path, stage)
    ext         = STAGE_DIRS[stage]
    ending      = ".%s" % ext
    try:
        files = filesystem.glob_dir(path, ending)
        for f in files:
            os.remove(os.path.join(path, f))
        log.debug(chlogger, {
            "name"      : __name__,
            "method"    : "prune",
            "path"      : ed_path,
            "feed"      : feed,
            "target_dir": p,
            "ending"    : ending,
            "removed_count": len(files),
            "message"   : "pruned target_dir",
            })
    except Exception as e:
        log.critical(chlogger, {
            "name"      : __name__,
            "method"    : "prune",
            "path"      : ed_path,
            "feed"      : feed,
            "target_dir": p,
            "ending"    : ending,
            "ERROR"     : "failed to prune target_dir",
            "exception" : str(e)
            })

def pre_reset(logger, feed, ed_path, stage):
    return os.path.join(ed_path, 'data', feed, STAGE_DIRS[stage])

def reset(logger, feed, ed_path, stage):
    chlogger = logger.getChild(__name__)
    p = pre_reset(logger, feed, ed_path, stage)
    try:
        if os.path.exists(p):
            shutil.rmtree(p)
            log.debug(chlogger, {
                "name"      : __name__,
                "method"    : "reset",
                "path"      : ed_path,
                "feed"      : feed,
                "target_dir": p,
                "message"   : "removed target_dir",
                })
    except Exception as e:
        log.critical(chlogger, {
            "name"      : __name__,
            "method"    : "reset",
            "path"      : ed_path,
            "feed"      : feed,
            "target_dir": p,
            "ERROR"     : "failed to remove target_dir",
            "exception" : str(e)
            })
    try:
        if not os.path.exists(p):
            os.makedirs(p)
            log.debug(chlogger, {
                "name"      : __name__,
                "method"    : "reset",
                "path"      : ed_path,
                "feed"      : feed,
                "target_dir": p,
                "message"   : "makedirs target_dir",
                })
    except Exception as e:
        log.critical(chlogger, {
            "name"      : __name__,
            "method"    : "reset",
            "path"      : ed_path,
            "feed"      : feed,
            "target_dir": p,
            "ERROR"     : "failed to makedirs target_dir",
            "exception" : str(e)
            })
    return p

def lines_in_file(f):
    try:
        with open(f, 'r') as x:
            lines = x.readlines()
            return len(lines)
    except:
        return 0

def src_files(logger, feed, ed_path):
    chlogger = logger.getChild(__name__)
    feed_dir    = os.path.join(ed_path, 'data', feed)
    src_dir     = os.path.join(feed_dir, 'src')
    src_files   = sorted(os.listdir(src_dir))
    log.debug(chlogger, {
            "name"      : __name__,
            "method"    : "src_files",
            "path"      : ed_path,
            "feed"      : feed,
            "feed_dir"  : feed_dir,
            "src_dir"   : src_dir,
            "src_files" : src_files
        })
    return src_files

def process_all_stages(logger, feed, ed_path):
    chlogger = logger.getChild(__name__)
    found_src_files = src_files(logger, feed, ed_path)
    if len(found_src_files) < 1:
        log.critical(chlogger, {
                "name"      : __name__,
                "method"    : "process_all_stages",
                "path"      : ed_path,
                "feed"      : feed,
                "src_files" : found_src_files,
                "ERROR"     : "No files found, nothing to process"
            })
        return
    log.debug(chlogger, {
            "name"      : __name__,
            "method"    : "process_all_stages",
            "path"      : ed_path,
            "feed"      : feed,
            "src_files" : found_src_files
        })
    for src_file in found_src_files:
        yield process_file(logger, feed, ed_path, src_file)

def process_stages(logger, feed, ed_path, stages):
    chlogger = logger.getChild(__name__)
    stage_files = [STAGE_PROCS[s] for s in stages]
    for sf in stage_files:
        if sf in src_files(logger, feed, ed_path):
            yield process_file(logger, feed, ed_path, sf)
        else:
            log.debug(chlogger, {
                    "name"      : __name__,
                    "method"    : "process_stages",
                    "path"      : ed_path,
                    "feed"      : feed,
                    "stage_file": sf,
                    "src_files" : src_files,
                    "ERROR"     : "stage_file not in src_files"
                })

def process_file(logger, feed, ed_path, src_file):
    chlogger    = logger.getChild(__name__)
    feed_dir    = os.path.join(ed_path, 'data', feed)
    rel_path    = os.path.join("src", src_file)
    cmd         = "%s %s" % (rel_path,  log.LOGGING_LEVEL_STRINGS[chlogger.getEffectiveLevel()])

    log.debug(chlogger, {
            "name"      : __name__,
            "method"    : "process_file",
            "path"      : ed_path,
            "feed"      : feed,
            "cmd"       : cmd
        })
    return runyield(cmd, feed_dir)


def archive_locally(logger, feed, ed_path, archivedir):
    chlogger = logger.getChild(__name__)
    try:
        archivedir1 = os.path.expanduser(archivedir)
        if archivedir1.startswith("/"):
            archivedire2 = archivedir1
        else:
            archivedir2 = os.path.join(ed_path, archivedir1)

        archive_name = os.path.join(archivedir2, feed)
        root_dir = os.path.expanduser(os.path.join(ed_path, 'data', feed))
        log.debug(chlogger, {
                "name"      : __name__,
                "method"    : "archive_locally",
                "path"      : ed_path,
                "feed"      : feed,
                "target_dir": archivedir2,
                "archive_name": archive_name,
                "root_dir"  : root_dir
            })
        return make_archive(archive_name, 'gztar', root_dir)
    except Exception as e:
        log.critical(chlogger, {
                "name"      : __name__,
                "method"    : "archive_locally",
                "path"      : ed_path,
                "feed"      : feed,
                "target_dir": archivedir2,
                "archive_name": archive_name,
                "root_dir"  : root_dir,
                "ERROR"     : "make archive failed",
                "exception" : str(e)
            })

def restore_locally(logger, feed, ed_path, archive):
    chlogger = logger.getChild(__name__)
    tf = tarfile.open(archive)
    feed_dir = os.path.join(ed_path, 'data', feed)
    if os.path.exists(feed_dir):
        log.critical(chlogger, {
                "name"      : __name__,
                "method"    : "restore_locally",
                "path"      : ed_path,
                "feed"      : feed,
                "archive"   : archive,
                "feed_dir"  : feed_dir,
                "ERROR"     : "Must delete the feed_dir before restoring."
                })
        return
    else:
        try:
            tf.extractall(os.path.join(ed_path, 'data', feed))
            return feed_dir
        except Exception as e:
            log.critical(chlogger, {
                    "name"      : __name__,
                    "method"    : "restore_locally",
                    "path"      : ed_path,
                    "feed"      : feed,
                    "archive"   : archive,
                    "feed_dir"  : feed_dir,
                    "ERROR"     : "Failed to restore archive to feed_dir",
                    "exception" : str(e)
                    })

def archive_to_s3(logger, feed, ed_path, service, bwlimit="100M"):
    """
    Archive feed dist to an S3 bucket.
    """
    chlogger    = logger.getChild(__name__)
    feed_dir    = os.path.join(ed_path, 'data', feed)
    dist_dir    = os.path.join(feed_dir, 'dist')
    s3_dir      = os.path.join('eap', 'energy-dashboard', 'data', feed)
    cmd = "rclone sync --bwlimit=%s --no-update-modtime --verbose %s/dist %s:%s" % (bwlimit, feed_dir, service, s3_dir)
    log.info(chlogger, {
            "name"      : __name__,
            "method"    : "archive_to_s3",
            "feed"      : feed,
            "path"      : ed_path,
            "service"   : service,
            "s3_dir"    : s3_dir,
            "cmd"       : cmd,
        })
    if not os.path.exists(dist_dir) \
            or not os.path.exists(os.path.join(dist_dir, 'zip')) \
            or not os.path.exists(os.path.join(dist_dir, 'db')):
        log.error(chlogger, {
                "name"      : __name__,
                "method"    : "archive_to_s3",
                "feed"      : feed,
                "path"      : ed_path,
                "dist_dir"  : dist_dir,
                "service"   : service,
                "s3_dir"    : s3_dir,
                "ERROR"     : "One of dist_dir|dist_dir/zip|dist_dir/db does not exist",
            })
        sys.exit(1)
    return runyield([cmd], feed_dir)

def s3_artifact_urls(logger, feed, ed_path, service):
    """
    Generate urls to access artifacts in S3 bucket.

    It'd be easy if we could simply 'rclone' from the S3 service and
    have the entire bucket replicated here. I've not had any luck with
    that approach.

    Here's the brute force solution. Use the state files,
    '[xml|sql|db|save]/state.txt', to direct the download operations.  

    Returns (url,target)
    :url: source
    :target: dest
    """
    chlogger    = logger.getChild(__name__)

    endpoints = {
            'digitalocean'  : 'sfo2.digitaloceanspaces.com',
            'wasabi'        : 's3.us-west-1.wasabisys.com'
    }
 
    def gen_url_target_tuples(stage):
        stages_idx  = STAGES.index(stage)
        out_dir     = DIRS[stages_idx]
        in_dir      = DIRS[stages_idx + 1]
        feed_dir    = os.path.join(ed_path, 'data', feed)
        state_file  = os.path.join(feed_dir, in_dir, 'state.txt')
        s3_dir      = os.path.join('eap', 'energy-dashboard', 'data', feed)
        with open(state_file, 'r') as artifacts:
            for a in artifacts:
                a       = a.rstrip()
                if stage == "download":
                    url     = "https://%s/%s/%s/%s" % (endpoints[service], s3_dir, out_dir, a)
                else:
                    # we don't upload the db directly, rather, we upload the pigz (.gz) file which is "name.gz"
                    url     = "https://%s/%s/%s/%s.gz" % (endpoints[service], s3_dir, out_dir, a)
                target  = os.path.join(feed_dir, out_dir, a)
                yield (url,target)

    tuples = []
    for stage in ['download', 'insert']:
        for t in gen_url_target_tuples(stage):
            tuples.append(t)
    return tuples



def restore_from_s3(logger, feed, ed_path, service):
    """
    Restore feed dist from an S3 bucket.

    It'd be easy if we could simply 'rclone' from the S3 service and
    have the entire bucket replicated here. I've not had any luck with
    that approach.

    Here's the brute force solution. Use the state files,
    '[xml|sql|db|save]/state.txt', to direct the download operations.  
    """
    chlogger    = logger.getChild(__name__)
    url_tuples = s3_artifact_urls(chlogger, feed, ed_path, service)
    try:
        for (url, target) in url_tuples:
            r = requests.get(url)
            if r.status_code == 200:
                with open(target, 'wb') as fd:
                    for chunk in r.iter_content(chunk_size=128):
                        fd.write(chunk)
                logger.info(chlogger, {
                    "name"      : __name__,
                    "method"    : "restore_from_s3",
                    "feed"      : feed,
                    "path"      : ed_path,
                    "service"   : service,
                    "url"       : url,
                    "target"    : target,
                    "message"   : "Restore succeeded",
                    })
            else:
                log.error(chlogger, {
                    "name"      : __name__,
                    "method"    : "restore_from_s3",
                    "feed"      : feed,
                    "path"      : ed_path,
                    "service"   : service,
                    "url"       : url,
                    "target"    : target,
                    "ERROR"     : "Failed to retrieve artifact from S3",
                    })
                target_parts = os.path.splitext(target)
                if target_parts[1] == ".gz":
                    subprocess.run("pigz -d %s" % target)
            # return downloaded urls
            yield url
    except Exception as e:
        log.critical(chlogger, {
                "name"      : __name__,
                "method"    : "restore_from_s3",
                "feed"      : feed,
                "path"      : ed_path,
                "service"   : service,
                "ERROR"     : "Failed to restore from S3",
                "exception" : str(e)
                })

def scanner(logger, feed, ed_path, xmlfile):
    chlogger    = logger.getChild(__name__)
    try:
        feed_dir    = os.path.join(ed_path(), 'data', feed)
        xml_dir     = os.path.join(feed_dir, 'xml')
        manifest    = os.path.join(feed_dir, 'manifest.json')
        with open(manifest, 'r') as f:
            obj = json.loads(f.read())
        if not xmlfile.startswith(xml_dir):
            xmlfile = os.path.join(xml_dir, xmlfile)
        logger.debug(chlogger, {
            "name"      : __name__,
            "method"    : "scanner",
            "feed"      : feed,
            "path"      : ed_path,
            "feed_dir"  : feed_dir,
            "xml_dir"   : xml_dir,
            "xml_file"  : xml_file,
            })
        with open(xmlfile, 'r') as f:
            return xmlparser.XML2SQLTransormer(f).parse().scan_types().scan_tables()
    except Exception as e:
        log.critical(chlogger, {
            "name"      : __name__,
            "method"    : "scanner",
            "feed"      : feed,
            "path"      : ed_path,
            "feed_dir"  : feed_dir,
            "xml_dir"   : xml_dir,
            "xml_file"  : xml_file,
            "ERROR"     : "Failed to parse and scan xml_file",
            "exception" : str(e)
            })

def generate_table_creation_ddl(logger, feed, ed_path, xmlfile):
    """
    Generate table creation SQL DDL from xmlfile.
    """
    return list(scanner(logger, feed, ed_path, xmlfile).ddl())

def generate_insertion_sql(logger, feed, ed_path, xmlfile, save):
    """
    Generate insertion SQL from xmlfile.
    """
    return list(scanner(logger, feed, ed_path, xmlfile).insertion_sql())

def get_latest_xml_file(logger, feed, ed_path, xml_dir):
    """
    Return the latest xml file (sorted by name, which includes a timestamp).
    May not actually be the latest due to vagaries in the names.

    If this is ever a problem then stat the files and use the lmod times.
    """
    xml_files   = sorted(list(fs.glob_dir(xml_dir, ".xml")))
    if len(xml_files) < 1:
        log.critical(logger, {
            "name"      : __name__,
            "method"    : "get_latest_xml_file",
            "feed"      : feed,
            "path"      : ed_path,
            "xml_dir"   : xml_dir,
            "ERROR"     : "Failed to parse xml_file",
            })
        return
    return xml_files[-1]

def create_and_save_ddl(logger, feed, ed_path, xmlfile, save):
    """
    Create the SQL DDL from an input xmlfile.
    Save the DDL to the manifest if save==True.
    Grab the latest xml file in the feed if xmlfile is None.
    """
    chlogger    = logger.getChild(__name__)
    feed_dir    = os.path.join(ed_path, 'data', feed)
    xml_dir     = os.path.join(feed_dir, 'xml')
    manifest    = os.path.join(feed_dir, 'manifest.json')

    if xmlfile is None:
        xmlfile = clifeed.get_latest_xml_file(logger, feed, ed_path, xml_dir)
    if not xmlfile.startswith(xml_dir):
        xmlfile = os.path.join(xml_dir, xmlfile)
    ddl = clifeed.generate_table_creation_ddl(logger, feed, path, xmlfile)
    if save:
        obj['ddl_create'] = ddl
        with open(manifest, 'w') as f:
            f.write(json.dumps(obj, indent=4, sort_keys=True))
    return ddl

def manifest_update(logger, feed, ed_path, field, value_str, value_int):
    """
    Update the manifest.field with the value provided.
    """
    feed_dir    = os.path.join(ed_path, 'data', feed)
    manifest    = os.path.join(feed_dir, 'manifest.json')

    value = value_str or value_int

    with open(manifest, 'r') as f:
        obj = json.loads(f.read())
    if value == None or value == "":
        obj.pop(field, None)
    else:
        obj[field] = value
    with open(manifest, 'w') as f:
        f.write(json.dumps(obj, indent=4, sort_keys=True))
